#!/usr/bin/env python

"""usage: asp_sync [options]

    -d  use other config dir than ~/.autoSyncPodder
    -h  this help
    -m  don't send mails - pipe all to stdout (default)

asp_download is part of the autoSyncPodder, which allows you
to fully automatic download your podcasts and sync them with your
mp3 player. After the installation and configuration you just need
to plug your mp3 into your pc and the data is transfered without any
interaction.

The asp_sync is in this setup responsible for syncing the pc with the
the mp3 player. 
"""

import sys
import os
import getopt
import statvfs
import shutil
import time
import beep

import config_manager
import lock

class Syncer:
    """ we created a class so we can reuse the code in easily in an
        other programs
    """
    _configManager = None
    _noMails = None

    def __init__(self, configManager, noMails):
        self._configManager = configManager
        self._noMails = noMails


    def _prepareSync(self):
        """ calculate which files need to copied/deleted and if
            all requirements (e.g. free space) are met
            this is an readonly operation
        """
        def getFiles(playerDir, syncInfoFilename):
            """ returns the files in 2 lists, the ones that exists
                and the others
            """
            filesExisting = []
            filesNotExisting = []
            for filenameRaw in file(syncInfoFilename):
                filename = filenameRaw.rstrip("\n")
                
                if os.path.isfile(os.path.join(playerDir, filename)):
                    filesExisting.append(filename)
                else:
                    filesNotExisting.append(filename)
            return filesExisting, filesNotExisting
        
        
        # get some base variables
        syncInfoFilename = self._configManager.getSyncInfoFilename()
        syncInfoInProcessFilename = self._configManager.getSyncInfoInProcessFilename()
        playerDir = self._configManager.getPlayerDir()
        syncDir = self._configManager.getSyncDir()
        
        # check if the mp3 player is prepared
        if not os.path.isdir(playerDir):
            raise IOError, "The directory %r does not exist, player plugged in and configurated?" % playerDir
        
        # first look for an old info file, if it exists something went
        # wrong at the last sync
        if os.path.isfile(syncInfoInProcessFilename):
            print "Something brock last time at syncing. Trying to fix it."
            filesExisting, filesNotExisting = getFiles(playerDir, syncInfoInProcessFilename)
        elif os.path.isfile(syncInfoFilename):
            filesExisting, filesNotExisting = getFiles(playerDir, syncInfoFilename)
        else:
            # running the first time or someone deleted our file, we won't delete anything
            # on the server
            filesExisting, filesNotExisting = [], []
        
        # build our sync list
        filesToSync = []
        filesToSyncSize = 0
        filesToDelete = []
        
        for filename in os.listdir(syncDir):
            # file got deleted on player?
            if filename in filesNotExisting:
                filesToDelete.append(filename)
                continue
            
            # file already on the player?
            if filename in filesExisting:
                continue
            
            # a new file
            filesToSync.append(filename)
            filesToSyncSize += os.stat(os.path.join(syncDir, filename)).st_size
            
        # enough space free on the stick?
        stat = os.statvfs(playerDir)
        free = stat[statvfs.F_BFREE] * stat[statvfs.F_FRSIZE]
        # one MB as reserve
        if free < filesToSyncSize + 1024**2:
            raise IOError, "Not enough space on the player left for copying (available: %d, required: %d" \
                            % (free, filesToSyncSize)
                    
        return filesToSync, filesToDelete, filesExisting

    def _renameSyncInfoFile(self):
        """ rename the sync Info File"""
        syncInfoFilename = self._configManager.getSyncInfoFilename()
        syncInfoInProcessFilename = self._configManager.getSyncInfoInProcessFilename()

        if os.path.isfile(syncInfoFilename):
            # (atomic function)
            os.rename(syncInfoFilename, syncInfoInProcessFilename)
        elif not os.path.isfile(syncInfoInProcessFilename):
            # create an empty file, if there is no inProcess File
            file(syncInfoInProcessFilename, "w").write("")
        
    def _copyFiles(self, filesToSync):
        """ copy the files onto the stick """
        syncDir = self._configManager.getSyncDir()
        playerDir = self._configManager.getPlayerDir()
        for filename in filesToSync:
            shutil.copy(os.path.join(syncDir, filename), playerDir)
    
    def _deleteFiles(self, filesToDelete):
        """ delete the files on the pc which have been deleted on the player """
        syncDir = self._configManager.getSyncDir()
        for filename in filesToDelete:
            fullPath = os.path.join(syncDir, filename)
            if os.path.exists(fullPath):
                os.unlink(fullPath)
        
    def _createNewSyncInfoFile(self, filesToSync, filesToKeep):
        """ create no sync Info File and delete old afterwards """        
        syncInfoFile = file(self._configManager.getSyncInfoFilename(), "w")
        for filename in filesToKeep + filesToSync:
            syncInfoFile.write(filename + "\n")
        
        os.unlink(self._configManager.getSyncInfoInProcessFilename())
        

    def sync(self):
        """ sync the pc with the mp3 player in a way that
            that we stay in an consistend state even if we are
            interrupted
        """        
        # 1. make sure all is ready and will work 
        filesToSync, filesToDelete, filesToKeep = self._prepareSync()
        
        # 2. rename the sync Info File
        self._renameSyncInfoFile()

        # 3. copy the files onto the stick
        self._copyFiles(filesToSync)
        
        # 4. delete the files on the pc which have been deleted on the player
        self._deleteFiles(filesToDelete)
        
        # 5. create no sync Info File and delete old afterwards
        self._createNewSyncInfoFile(filesToSync, filesToKeep)

# Print usage message and exit
def usage(*args):
    sys.stdout = sys.stderr
    print __doc__
    print 50*"-"
    for msg in args: print msg
    sys.exit(2)
    
# Main program: parse command line and start processing
def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'd:hm')
    except getopt.error, msg:
        usage(msg)

    noMails = False
    configDir = None
    
    for o, a in opts:
        if o == '-h':
            print __doc__
            sys.exit()
        if o == '-d':
            configDir = a
        if o == '-m':
            noMails = True
    
    myConfigManager = config_manager.ConfigManager(configDir)

    # first lock other instances of me
    if not lock.test_and_lock(myConfigManager.getSyncLockFile()):
        #FIXME: send mail, as this is an error!
        print "An other instance of asp_sync is already running! Aborting"
        sys.exit(2)

    time.sleep(5)
    mySyncer = Syncer(myConfigManager,noMails)
    mySyncer.sync()
    lock.unlock(myConfigManager.getSyncLockFile())
    
    beep.beep(440, 63, 1)

if __name__ == '__main__':
    main()
